'' /////////////////////////////////////////////////////////////////////////////
'' Graphics Renderer for Donkey Kong Remake
'' (C) Steve Waddicor
'' //////////////////////////////////////////////////////////////////////////////

CON
  PARAM_COUNT   = 11
  NUMBER_OF_STRINGS = 32

OBJ
'  tv    : "sw_dk_tv_drv_022.spin"

VAR
  long  cog

VAR
    ' Renderer parameter block
    ' Each renderer copies these values to it's own cog_memory cache
    ' when it's got a little spare time, several times per frame
    ' so these values can be altered on the fly.
    long    rend_cog_number        'each cog gets a different number, starting with zero     write-only
    long    rend_cog_count         'total number of cogs                                     write-only
    long    rend_scanline_req      'a pointer to the request to put a line in the scan buffer   write-only
    long    rend_scanline_buffer   'hub memory where renderer puts scanline when requested write-only
    long    rend_text_location_table
    long    rend_text_color_table
    long    rend_text_string_table
    long    rend_scanxsize
    long    rend_scanysize
    long    rend_mode
    long    rend_tvstatus_ptr
    'End of renderer parameter block

    ' TV parameter block
'    long    tv_scanline_request
'    long    tv_scanline_buffer
'    long    tv_status
    ' End of TV parameter block

    long    string_locations[32]
    long    end
    'string_colors
    '0 = invisible
    byte    string_colors[32]
    'string_ptrs
    '0 = this string unused.  OK to reuse.
    word    string_ptrs[32]

PUB Start(aCogs,TvParamsPtr) | i

    repeat i from 0 to NUMBER_OF_STRINGS-1
        string_locations[i]:=$FFFF_FFFF    
        string_colors[i] := 0
        string_ptrs[i] := 0
    end:=$FFFF_FFFF
    string_locations[0]:=$3800_6000
    string_colors[0] := $8e
    string_ptrs[0] := @hi
    
    rend_cog_count := aCogs
    rend_scanline_req := long[TvParamsPtr+(14*4)]
    rend_scanline_buffer := long[TvParamsPtr+(12*4)]
    rend_text_location_table := @string_locations
    rend_text_color_table := @string_colors
    rend_text_string_table := @string_ptrs
    rend_scanxsize := long[TvParamsPtr+(4*4)]*4
    rend_scanysize := long[TvParamsPtr+(5*4)]*16
    rend_mode := long[TvParamsPtr+(3*4)]
    rend_tvstatus_ptr := TvParamsPtr

    repeat i from 0 to aCogs-1
        rend_cog_number := i
        cognew(@entry,@rend_cog_number)
        'Wait to allow this cog to get the cog number param from the block before changing it.
        repeat 1000

'    tv_scanline_request := @scanline_req
'    tv_scanline_buffer := @hub_scanline_buffer
'    tv.start(@tv_scanline_request)
    'return address of tv_status to caller
'    long[aTvStatusPtr]:=@tv_status

PUB NewString(aId,aX,aY,aString,aChroma,aLuma):rHandle
    {{
    aId:            An identifier for referring to the string.  This must be in
                    the range 0 =< aId < NUMBER_OF_STRINGS.  Strings with
                    higher IDs display in front of strings with lower IDs.
                    
    word aX,aY:     Location of bottom left of string.  Coordinates have their
                    origin at the top left of the screen and y-axis increases
                    down the screen.  Coodinates are in fixed point form, with
                    4 bits for the fractional part.  Thus:
                        $0000,$00F0 will display a string at the top left of the screen.
                        $0000,$0100 will display one pixel lower down the screen.
                        $0000,$0DF0 will display on the bottom line of the screen.

    string aString: The string to display

    byte    aColor: The color to display the text in.  This is a hi-color value.
                    The special value of 0 will make the text invisible.
    }}
    if aId=>NUMBER_OF_STRINGS
        return -1
    string_locations[aId]:=(aX<<16)+aY
    string_ptrs[aId]:=aString               
    string_colors[aId]:=((aChroma&$0f)<<4) + aLuma&$3 + 11

PUB DeleteString(aId)
    string_ptrs[aId]~

PUB DeltaStringPos(aId,aDeltaX,aDeltaY)
    string_locations[aId]+=aDeltaX<<16+aDeltaY
    
PUB DeltaStringColor(aId,aDeltaChroma,aDeltaLuma) | chroma, luma
    chroma:=string_colors[aId]>>4
    luma  :=(string_colors[aId] - 3)&%11
    chroma+=aDeltaChroma
    luma+= aDeltaLuma
    luma #>= 0
    luma <#= 3
    string_colors[aId]:=(chroma&$0f)<<4 + luma + 11

PUB MoveStringTo(aId,aX,aY)
    string_locations[aId]:=(aX<<16)+aY
    
DAT
hi      byte  "Fantasia!",0    
DAT

'******************************
'* Assembly language renderer *
'******************************
                        org
'
'
' Entry
'
entry
                        rdlong  cog_rend_cog_number,PAR
                        mov     current_line,cog_rend_cog_number  'start rendering from the top
                        ' Load parameters
                        mov     r0,PAR
                        movd    :load_dest1,#renderer_params+1
                        mov     count,#param_count-1
:load1                   add     r0,#4
                        '       ___
:load_dest1              rdlong  0-0,r0
                        '       ^^^ iterates through renderer_params array.
                        add     :load_dest1,inc_dest
                        djnz    count,#:load1
                        test    cog_rend_mode,#%0010    wz      'if interlaced
               if_nz     add     current_line,cog_rend_cog_number

do_line

'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

                        mov     r0,cog_rend_scanline_buffer
                        add     r0,#256
                        add     r0,#256
                        add     r0,current_line
                        rdbyte  r1,r0
                        mov     r0,r1
                        rol     r0,#8
                        or      r0,r1
                        mov     r1,r0
                        rol     r0,#16
                        or      r0,r1
                        
                        movd    :clear_line,#cog_scanline_buffer
                        mov     count,cog_rend_scanxsize '#64
:clear_line             mov     0-0,black
':clear_line             mov     0-0,r0
                        add     :clear_line,inc_dest
                        djnz    count,#:clear_line

'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

                        mov     hub_location_tab_ptr,cog_rend_text_location_table
                        mov     hub_color_tab_ptr,cog_rend_text_color_table
                        mov     hub_string_tab_ptr,cog_rend_text_string_table
:string_loop            rdlong  location,hub_location_tab_ptr
                        cmp     location,end_of_location_array wz
              if_e      jmp     #:end_string_table
              
                        mov     y,location
                        shr     y,#4
                        and     y,FFF
                        mov     char_line,current_line
                        sub     char_line,y wz,wc
              if_b      jmp     #:next_string

                        cmp     char_line,#32 wz,wc
              if_ae     jmp     #:next_string
                        shl     char_line,#2
              
                        rdbyte  color,hub_color_tab_ptr wz
              if_z      jmp     #:next_string

                        mov     pix_ptr,#cog_scanline_buffer
                        mov     x,location
                        shr     x,#22     
                        add     pix_ptr,x

                        rdword  string_ptr,hub_string_tab_ptr
:char_loop              rdbyte  char_ptr,string_ptr wz
              if_z      jmp     #:next_string
                        mov     offset,char_ptr
                        
                        shr     char_ptr,#1
                        shl     char_ptr,#7
                        add     char_ptr,char_set_start
                        add     char_ptr,char_line

                        rdlong  char,char_ptr wz

                        shr     offset,#1 wc
              if_c      shr     char,#1

                        mov     nybble_count,#4

:next_nybble            cmps    pix_ptr,#cog_scanline_buffer wz,wc
              if_b      jmp     #:skip_off_screen
                        cmps    pix_ptr,#end_cog_scanline_buffer wz,wc
              if_ae     jmp     #:skip_off_screen

                        movs    :load1,pix_ptr
                        movd    :save1,pix_ptr
:load1                  mov     pixels,0-0

                        mov     pixel_count,#4
:next_pixel             shr     char,#2 wc
              if_c      and     pixels,ffffff00
              if_c      or      pixels,color
                        ror     pixels,#8
                        djnz    pixel_count,#:next_pixel

:save1                  mov     0-0,pixels

:skip_off_screen        add     pix_ptr,#1 
                        djnz    nybble_count,#:next_nybble
                        
                        add     string_ptr,#1
                        jmp     #:char_loop
                        
:next_string            add     hub_string_tab_ptr,#2
                        add     hub_color_tab_ptr,#1
                        add     hub_location_tab_ptr,#4
'                        jmp     #:string_loop
                        call    #checktv
              if_b      jmp     #:string_loop
                        
:end_string_table
                        
'@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

                        ' We've finished rendering the line and we didn't have a hissy fit panic.
                        ' We need to wait here for the line we just prepared to be requested by the TV driver
wait_for_the_request
                        call    #checktv
              if_b      jmp     #wait_for_the_request

                        ' Our line has been requested so copy it to hub scan line buffer
start_tv_copy           movs    :nextcopy, #cog_scanline_buffer
                        mov     r0, cog_rend_scanline_buffer
                        mov     count,cog_rend_scanxsize ' #80              ' length of visible scanline buffer in longs
:nextcopy               mov     r1,cog_scanline_buffer
                        add     :nextcopy, #1
                        wrlong  r1, r0
                        add     r0, #4
                        djnz    count, #:nextcopy

                        ' We've handed over the scanline to the TV driver, so do another.
allocate_next_job       add     current_line,cog_rend_cog_count 'we only do every Nth line
                        test    cog_rend_mode,#%0010    wz      'if interlaced
              if_nz     add     current_line,cog_rend_cog_count
              
                        ' Degrade gracefully when there's too much work to do.
                        call    #checktv
              if_ae     jmp     #allocate_next_job

                        cmp     current_line,cog_rend_scanysize wc,wz
              if_b      jmp     #do_line

                        test    cog_rend_mode,#%0010    wz      'if interlaced
              if_nz     jmp     #start_interlaced
                        mov     current_line,cog_rend_cog_number
                        jmp     #do_line
start_interlaced
                        mov     r0,cog_rend_tvstatus_ptr
                        add     r0,#1
                        rdbyte  current_line,r0
                        and     current_line,#1
                        add     current_line,cog_rend_cog_number
                        add     current_line,cog_rend_cog_number
                        jmp     #do_line
              
'-------------------------------------------------------------------
checktv                 ' Tf we're currently rendering a low numbered line when a high one is being requested
                        ' then that's OK because we wrapped around.
                        rdlong  scanline_req,cog_rend_scanline_req

                        cmp     scanline_req,cog_rend_scanysize wc,wz
              if_b      jmp     #:checkit
              
                        cmp     current_line,#14 wc,wz
              if_b      cmp     scanline_req, current_line wz
                        'b = normal
                        'z = doit
                        'a = err
              if_be     jmp     #checktv_ret

:checkit                cmp     scanline_req, current_line wz, wc    'set carry if request is less than current.
                        'b = normal
                        'z = doit
                        'a = err

checktv_ret             ret
'-------------------------------------------------------------------

' Initialized data
black                   long    $02_02_02_02
end_of_location_array   long    $FFFF_FFFF
char_set_start          long    $8000
ffffff00                long    $FFFFFF00
fff                     long    $FFF
inc_dest                long    1 << 9 << 0

hub_location_tab_ptr    long    0
hub_color_tab_ptr       long    0
hub_string_tab_ptr      long    0
char_ptr                long    0
char                    long    0
location                long    0
x                       long    0
y                       long    0
color                   long    0
pix_ptr                 long    0
pixels                  long    0
string_ptr              long    0
offset                  long    0
char_line               long    0
nybble_count            long    0
pixel_count             long    0
' End of scanline buffer
scanline_req            long    0

DAT 'Uninitiaised data - always keep it at the end of file to avoid nasty gotcha with RES.

renderer_params  
cog_rend_cog_number             res 1 'each cog gets a different number, starting with zero     write-only
cog_rend_cog_count              res 1 'total number of cogs                                     write-only
cog_rend_scanline_req           res 1 'a pointer to the request to put a line in the scan buffer   write-only
cog_rend_scanline_buffer        res 1 'hub memory where renderer puts scanline when requested write-only
cog_rend_text_location_table    res 1
cog_rend_text_color_table       res 1
cog_rend_text_string_table      res 1
cog_rend_scanxsize              res 1
cog_rend_scanysize              res 1
cog_rend_mode                   res 1
cog_rend_tvstatus_ptr           res 1
    
' Cog scanline buffer = 256 pixels, 8 bits per pixel = 2048 bits = 256 bytes = 64 longs
cog_scanline_buffer     res     88
end_cog_scanline_buffer

current_line            res     1
r0                      res     1
r1                      res     1
r2                      res     1
count                   res     1
_mode                   res     1

                        fit